function varargout = genmpc(cfg)
% Generic MPC (online planning) control algorithm
%   VARARGOUT = GENMPC(CFG)
% Inputs:
%   CFG             - structure with fields as commented in the code below
%           can also be given as a string, see str2cfg
% Outputs:
%   none, in run mode
%  [HIST, FIGH]     - history structure and figure handle, in plottrial mode
%   FIGH            - in ploth, plotdepth, plotn modes
%
% Supports using reference model (not integrated in plotting) 

if nargin < 1, cfg = struct(); end;

% -----------------------------------------------
% Process configuration structure

% default config
% script config
CFG.model_params = {};              % extra parameters for problem calling in 'model' mode
CFG.run = 0;                        % run control
CFG.plottrial = 0;                  % show some trajectories
CFG.ploth = 0;                      % plot policy: can be an action index if multidimensional action
CFG.plotdepth = 0;                  % plot depths of computed OP trees
CFG.plotn = 0;                      % plot budget taken for each OP tree (useful when run to set depth)
CFG.problem = '';                   % what problem to solve
CFG.datadir = '';                   % data dir
CFG.datafile = '';                  % save data to file
CFG.online = 1;                     % 1: control online / 0: find control for given states, offline
CFG.gamma = 0.98;                   % discount factor
CFG.planalg = 'opd';                % planning algorithm
CFG.planparams = [];                % planning algorithm parameters as a structure
CFG.randseed = [];                  % set the random seed (for repeatable results)
% for online control:
CFG.trials = 1;                     % # trials (control episodes)
CFG.maxsteps = 100;                 % max # steps in a trial
CFG.reset = 'rand';                 % trial reset
CFG.actiontype = 'one';             % 'one' action only, or 'seq'uence of actions
CFG.maxseqlen = Inf;                % maximum length of action sequence to apply (whole seq if shorter)
    % if maxseqlen >= 1, then this exact amount of steps is applied
    % if maxseqlen < 1, then a fraction ceil(maxseqlen * length_of_maximal_sequence) is applied    
    % note the structure of the planinfo and plandata histories changes when the actiontype is seq
    % also note that batch exps render Inf as "1" due to interpreting them as maximal fractionary length
    % (so setting depth and then maxseqlen=Inf does not produce the expected result)
% for offline computation of controls:
CFG.X = [];                         % set of states where to find actions
% stats config
CFG.saveplandata = [];              % remember the data created by planning algorithm -- memory-intensive!
% display config
CFG.verb = 0;                       % verbosity: the higher, the more detailed the messages displayed
CFG.visualize = 0;                  % for online, visualization level (0 = none, 1 = trial-level, 2 = step-level)
CFG.viscfg = struct;                % visualization config options
CFG.trialdisp = 10;
CFG.sampledisp = .05;
% figure config
CFG = setfigprop(CFG, 'addfields'); % figure config fields

% List of fields that define the problem and therefore may NOT be overwritten on load
KEEPFIELDS = {'problem', 'gamma', 'online', 'actiontype', 'planalg', 'planparams', 'X'};

% Install function defaults for everything else
cfg = parseconfig(cfg, CFG);

if ~cfg.run,        % load data file
    loadvars = {'cfg', 'model', 'planinfoh'};
    if cfg.plottrial,
        loadvars = {loadvars{:}, 'X', 'U', 'R', 'p', 'q'};
    elseif cfg.ploth,
        loadvars = {loadvars{:}, 'X', 'U', 'p', 'q'};
    elseif cfg.plotdepth || cfg.plotn,
        loadvars = {loadvars{:}, 'U', 'planinfoh', 'plansteps'};
    end;
    % make sure that cfg is not overwritten
    cfg1 = cfg; kf = KEEPFIELDS;
    load(cfg.datafile, loadvars{:});
    % Overwrite problem-defining fields from loaded config, keep the rest as in the (current) cfg1;
    % the result becomes the new config
    % backward compatibility with older versions that did not have certain fields
    if ~isfield(cfg, 'online'),    cfg.online = 1; end;  
    if ~isfield(cfg, 'X'),         cfg.X = []; end;  
    if ~isfield(cfg, 'actiontype') cfg.actiontype = 'one'; end;
    cfg = copyfields(cfg, cfg1, kf);
    KEEPFIELDS = kf; clear cfg1 kf;
    dispx(['Data loaded from [' cfg.datafile '].'], cfg.verb, 1);
end;

% Echo config
dispx('GENMPC called with the following configuration:', cfg.verb, 1);
dispx(cfg, cfg.verb, 1, 'cfg', 'config');
planparams = cfg.planparams; dispx(planparams, cfg.verb, 1, 'planparams', 'config'); clear planparams;

% get environment (Matlab, hardware) info
cfg.envinfo = getenvx;

% -----------------------------------------------
% Main algorithm
if cfg.run,
    
    % create model, get its dimensionality data
    model = feval(cfg.problem, 'model', cfg.model_params{:});
    
    p = model.p; q = model.q;
    
    % error checks
    if ~cfg.online && model.userefmodel, error('GENMPC cannot use a reference model in offline mode'); end;
    if cfg.actiontype(1) == 's',
        if ~cfg.online, error('GENMPC cannot apply action sequences in offline mode'); end;
        if cfg.maxseqlen < 0, error('Invalid maxseqlen value'); end;
        if strcmp(cfg.planalg, 'opss') && isfinite(cfg.maxseqlen), error('Limiting seq length in OPSS is not welldefined'); end;
    end;
    % init planning config, copy the configuration fields
    plancfg.gamma = cfg.gamma;
    plancfg = copyfields(cfg.planparams, plancfg);
    plancfg.actiontype = cfg.actiontype;
    plancfg.maxseqlen = cfg.maxseqlen;          % in case sequences are required
    
    % init random generator if configured
    if ~isempty(cfg.randseed), rand('twister', cfg.randseed); end;
    
    % if X is string identifier, get actual states from the problem
    if ischar(cfg.X), cfg.X = feval(cfg.problem, 'X0', cfg.X); end;
    % process state set
    X = cfg.X; 
    if iscell(X), X = flat(X); end;
    % if "smart" reset is used, copy states to reset structure 
    % -- unless overridden by existing setting
    if isstruct(cfg.reset) && ~isfield(cfg.reset, 'x0'), cfg.reset.x0 = X; end;
    
    if cfg.online,
        % ONLINE control, along trajectories of the system
        
        % init histories etc.
        X = NaN(p, cfg.maxsteps+1, cfg.trials);
        U = NaN(q, cfg.maxsteps+1, cfg.trials);
        R = NaN(1, cfg.maxsteps+1, cfg.trials);
        if model.userefmodel, REFX = NaN(model.refp, cfg.maxsteps+1, cfg.trials); end;
        % note in sequence mode, not all of these will be populated
        if cfg.saveplandata, plandatah = cell(cfg.maxsteps, cfg.trials); end;
        % this flags the steps where planning has actually been run
        % (which in sequence mode will not be all the steps)
        planflags = false(cfg.maxsteps, cfg.trials);
        % stores sequence information in a more user-friendly mode, as vectors [k len] for each
        % sequence (one mnatrix of such vectors for each trial)
        if cfg.actiontype(1) == 's',
            plansteps = cell(1, cfg.trials);
        end;
        % online error estimates for each trial -- TODO implement
        % ERR = NaN(1, cfg.trials);
        
        trial = 1;
        timestat.run = 0;
        % note in sequence mode, not all of these will be populated
        timestat.planning = nan(cfg.maxsteps, cfg.trials);

        % init visualization config if needed
        if cfg.visualize,
            vcfg = cfg.viscfg;
            vcfg.gview = [];
        end;

        % -----------------------------------------------
        % Run control
        dispx('Performing GENMPC online control...', cfg.verb, 0);

        t = cputime;
        while trial <= cfg.trials,       % main loop

            % initialize trial
            k = 1; terminal = 0;
            % initial state
            % if "smart" reset is used, copy trial index as it may be needed
            if isstruct(cfg.reset), cfg.reset.trial = trial; end;
            if model.userefmodel,
                [X(:, 1, trial), REFX(:, 1, trial)] = feval(cfg.problem, 'reset', model, cfg.reset);
            else
                X(:, 1, trial) = feval(cfg.problem, 'reset', model, cfg.reset);
            end;

            timestat.run = timestat.run + (cputime - t);
            % visualize initial state
            if cfg.visualize >= 2,
                vcfg.mpcstep = 1;
                vcfg.mpctrial = 0;
                vcfg.k = k - 1;
                vcfg.trial = trial;
                vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
                [figh vcfg.gview] = feval(model.visualizefun, vcfg);
            end;
            t = cputime; if cfg.actiontype(1) == 's', replan = true; end;

            while k <= cfg.maxsteps && ~terminal,        

                % choose action or action sequence using the configured algorithm
                t2 = cputime;
                
                if cfg.actiontype(1) == 'o',    
                    % plan one action per step
                    plancfg.k = k;      % send k to the planning algorithm
                    if model.userefmodel, plancfg.refx = REFX(:, k, trial); end;
                    if ~cfg.saveplandata,   % save some time by not building the plandata
                        [U(:, k, trial), planinfo] = feval(cfg.planalg, X(:, k, trial), model, plancfg);
                    else
                        [U(:, k, trial), planinfo, plandata] = feval(cfg.planalg, X(:, k, trial), model, plancfg);
                    end;
                    planflags(k, trial) = true;
                    timestat.planning(k, trial) = cputime - t2;
                elseif cfg.actiontype(1) == 's',
                    % sequence of actions -- plan only if needed (first step, or previous sequence ran out)
                    if replan,
                        plancfg.k = k;
                        if model.userefmodel, plancfg.refx = REFX(:, k, trial); end;
                        [useq, planinfo, plandata] = feval(cfg.planalg, X(:, k, trial), model, plancfg);
                        actionsleft = size(useq, 2);
                        planflags(k, trial) = true;
                        timestat.planning(k, trial) = cputime - t2;
                    end;
                    if strcmp(cfg.planalg, 'opss'),      % we must actually build the seq from a closedloop tree
                        if replan, 
                            node = 1; useq = []; % root = single outcome, initially
                        else
                            % outcomenodes holds all possible stochastic outcomes of last action
                            ichild = findflat(X(:, k, trial), plandata.Tx(:, outcomenodes), 1, 'first');
                            node = outcomenodes(ichild);
                            if isempty(node), error('No outcome matches'); end;
                        end;
                        % choose the action with the best util (nu, lower bound) value
                        [utilmax, jmax] = max(plandata.Tutil(:, node));
                        U(:, k, trial) = plancfg.U(:, jmax);
                        useq(:, end+1) = plancfg.U(:, jmax);
                        % children of this action are the new outcome nodes
                        outcomenodes = plandata.Tchildren(jmax, :, node);
                        outcomenodes = outcomenodes(outcomenodes > 0);  % ignore uninitialized children
                        % if any are leaves (<=> all are leaves, given the algo), we must replan
                        replan = any(plandata.Tleaf(outcomenodes));
                    else
                        U(:, k, trial) = useq(:, end-actionsleft+1);
                        actionsleft = actionsleft - 1;
                        replan = (actionsleft == 0);
                    end;
                end;

                % apply the action
                if model.userefmodel,
                    [X(:, k+1, trial), R(1, k+1, trial), terminal, dummypplus, REFX(:, k+1, trial)] = ...
                        feval(model.fun, model, X(:, k, trial), U(:, k, trial), [], REFX(:, k, trial));
                else
                    [X(:, k+1, trial), R(1, k+1, trial), terminal] = ...
                        feval(model.fun, model, X(:, k, trial), U(:, k, trial));
                end;

                % update stats
                timestat.run = timestat.run + (cputime - t);
                if planflags(k, trial), 
                    planinfoh(k, trial) = planinfo;
                    if cfg.saveplandata, plandatah{k, trial} = plandata; end;
                end;

                % log
                if cfg.verb >= 3,
                    if planflags(k, trial), message = planinfo.message;
                    else message = 'continuing with action sequence';
                    end;
                    disp(sprintf('k=%d: x=%s u=%s (%s)', k, ...
                        num2strx(X(:, k, trial), 'vector'), num2strx(U(:, k, trial), 'vector'), message));
                    drawnow;
                end;

                % visualization
                if cfg.visualize >= 2,
                    vcfg.mpcstep = 1;
                    vcfg.mpctrial = 0;
                    vcfg.k = k;
                    vcfg.trial = trial;
                    vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
                    [figh vcfg.gview] = feval(model.visualizefun, vcfg);
                end;

                % finalize loop
                t = cputime; 
                k = k + 1;
            end;

            % update stats
            timestat.run = timestat.run + (cputime - t);

            % planning steps
            if cfg.actiontype(1) == 's',
                planind = find(planflags(:, trial));
                % first col of plansteps: the indices where planning was done
                % second col: the length of the applied plans
                plansteps{trial} = [planind(:) [diff(planind(:)); k - planind(end)]];
            end;

            % visualization
            if cfg.visualize >= 1,
                vcfg.mpctrial = 1;
                vcfg.mpcstep = 0;
                vcfg.trial = trial;
                vcfg.ktotal = k - 1 + cfg.maxsteps * (trial - 1); % this only works in continuing tasks
                [figh vcfg.gview] = feval(model.visualizefun, vcfg);
            end;

            % console feedback of algorithm progress
            if ~mod(trial, cfg.trialdisp), 
                dispx(sprintf('Trial #%d completed.', trial), cfg.verb, 2);
            end;

            % start counting time again, increment iteration counter
            t = cputime;
            trial = trial + 1;
        end;        % trials loop

        dispx(sprintf('trials=%d exhausted. Algorithm stopped.', cfg.trials), cfg.verb, 0);

        % finalize visualizer
        if cfg.visualize,
            vcfg.mpctrial = 0;
            vcfg.mpcstep = 0;
            vcfg.finalize = 1;
            [figh vcfg.gview] = feval(model.visualizefun, vcfg);
        end;

        % end online control branch
    else
        % OFFLINE control
        % initialize histories etc.
        N = size(X, 2);
        U = NaN(q, N);
        if cfg.saveplandata, plandatah = cell(1, N); end;

        timestat.run = 0;
        timestat.planning = nan(1, N);

        % -----------------------------------------------
        % Run control
        dispx('Performing GENMPC offline... ', cfg.verb, 0);
        dispx([8 sprintf('states processed: %3d%%', 0)], cfg.verb, 2);

        for i = 1:N,
            % choose action using the configured algorithm
            t = cputime;
            
            if cfg.saveplandata, 
                [U(:, i), planinfo, plandata] = feval(cfg.planalg, X(:, i), model, plancfg);
            else
                [U(:, i), planinfo] = feval(cfg.planalg, X(:, i), model, plancfg);
            end;

            % update stats and output console log
            timestat.planning(i) = cputime - t;
            planinfoh(i) = planinfo;
            if cfg.saveplandata, plandatah{i} = plandata; end;
            if cfg.verb >= 2 && mod(i/N, cfg.sampledisp) <= 1/N,    % display progress inline
                disp([8 8 8 8 8 sprintf('%3d%%', fix(i/N*100))]);
            end;
        end;
        % update stats
        timestat.run = sum(timestat.planning);
        dispx([8 sprintf('done in %.2fs.', timestat.run)], cfg.verb, 0);
    end;    % IF online
    
end;

% Plot a trial (only makes sense if algorithm ran online)
if cfg.plottrial && cfg.online,
    Ns = find(~isnan(U(1, :, cfg.plottrial)), 1, 'last');
    hist.t = (0:Ns) * model.Ts; 
    hist.x = X(:, 1:Ns+1, cfg.plottrial); 
    hist.u = U(:, 1:Ns+1, cfg.plottrial); 
    hist.r = R(1, 1:Ns+1, cfg.plottrial);
    % hist.refx = REFX(:, 1:Ns+1, cfg.plottrial);   % not being loaded in current implem
    hist.R = discreturn(cfg, hist.r, Ns, 0);    % note the 0 should in fact be the "terminal" flag
    mpfigh = [];
    % if the model provides a plot function, use it
    if isfield(model, 'plotfun'),  
        plotfun = model.plotfun;
    elseif isfield(model, 'problem') && ~isempty(feval(model.problem, 'plotfun')),
        plotfun = feval(model.problem, 'plotfun');
    else
        plotfun = '';
    end;
    if ~isempty(plotfun),
        mpcfg = struct;
        mpcfg.hist = hist;
        mpcfg.trajectory = 1; % NOTE this will differ for various plot functions...
        mpcfg.traj = hist;    % this is for dcplot
        mpcfg = copyfields(cfg, mpcfg, fieldnames(setfigprop(struct, 'addfields')));
        mpfigh = feval(plotfun, mpcfg);
    end;
    % otherwise, or if the plot function produced nothing, make a generic plot
    if ~isempty(mpfigh), figh = mpfigh;
    else figh = plothistory(hist); 
    end;
    setfigprop(cfg);
    % save if requested
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    
    varargout = {hist, figh};
end;

% Plot depth / budget info for the first trial
if (cfg.plotdepth || cfg.plotn) && cfg.online,
    trial = 1;
    Ns = find(~isnan(U(1, :, trial)), 1, 'last');
    hist.t = (0:Ns) * model.Ts; 
    figh = figure; 
    ps = plansteps{trial}(:, 1);
    if cfg.plotdepth,
        scatter(ps' * model.Ts, [planinfoh(:, 1).depth], [], 'k', 'filled');
        if any([planinfoh(:, 1).depth] ~= plansteps{trial}(:, 2)'),
            hold on;
            scatter(ps' * model.Ts, plansteps{trial}(:, 2), [], 'k', 'x');
            legend('Planning depth', 'Length of applied subsequences');
        end;
        ylabel('d');
    elseif cfg.plotn,
        scatter(ps' * model.Ts, [planinfoh(:, 1).n], [], 'k', 'filled');
        set(gca, 'yscale', 'log');
        ylabel('n');
    end;
    
    xlabel('t [s]');
    set(gca, 'xlim', minmax(hist.t));
    
    setfigprop(cfg);
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    varargout = {figh};
end;

% Plot policy -- only makes sense if algorithm ran offline
if cfg.ploth && ~cfg.online,
    % X must be a cell array of grids
    if model.p ~= 2 || model.q ~= 1, error('sol mode only works for 2x1 problems'); end;
    
    figh = figure;
    scatter(X(1, :), X(2, :), 25, U(cfg.ploth, :), 'filled', 'MarkerEdgeColor', 'k', 'LineWidth', .1);
    set(gca, 'xlim', [-model.maxx(1) model.maxx(1)]);
    set(gca, 'ylim', [-model.maxx(2) model.maxx(2)]);
    set(gca, 'clim', [-model.maxu model.maxu]);
    xlabel('x_1'); ylabel('x_2'); title('h(x_1, x_2)');
    colormap(gray);
    
    setfigprop(cfg);
    saveplot(figh, [cfg.savedir cfg.savefig], cfg.plottarget);
    varargout = {figh};    
end;



% -----------------------------------------------
% Backup data

if cfg.run && ~isempty(cfg.datafile),
    % save into the same directory as the problem
    if isempty(cfg.datadir),
        datadir = fileparts(which(cfg.problem));
    else
        datadir = cfg.datadir;
        if (datadir(end) == '\'), datadir = datadir(1:end-1); end;
    end;
    cfg.datafile = [datadir '\' cfg.datafile];
    save(cfg.datafile);
    dispx(['Data was saved to [' cfg.datafile '].'], cfg.verb, 1);
end;


% END qi() RETURNING varargout =================================================================

% % code for handling sequence length here
%                         % debug output: fprintf('k = %d, received sequence of length %d... ', k, size(useq, 2));
%                         % if sequence is longer than desired, truncate it
%                         if ~isfinite(cfg.maxseqlen),
%                             % maximal length sequences -- no change
%                         elseif cfg.maxseqlen >= 1,
%                             % absolute maximum number of steps
%                             if size(useq, 2) > cfg.maxseqlen, useq = useq(:, 1:cfg.maxseqlen); end;
%                         elseif cfg.maxseqlen > 0 && cfg.maxseqlen < 1,
%                             % fraction of returned sequence length
%                             useq = useq(:, 1:ceil(size(useq, 2) * cfg.maxseqlen));
%                         else
%                             error('Invalid maxseqlen value: %f', cfg.maxseqlen);
%                         end;
%                         % debug output: fprintf('applying subsequence of length %d\n', size(useq, 2));


%         % in sequence mode, process planning flags to produce an easy-to-use table:
%         % [k len] where k = step where planning was run, len = length of planned sequence
%         plansteps = cell(1, cfg.trials);
%         for i = 1:cfg.trials,
%             if any(isnan(X(:, :, i)), error('plansteps computation does not support early termination'); end;
%             plansteps{i} = [find(planflags(:, i)) ...
%                 diff([find(planflags(:, i)) cfg.maxsteps+1])];
%             % note last sequence may actually have been longer
%         end;
